%{
open Control

let extensions = [".tar.gz"; ".tgz"; ".tar.bz2"; ".tbz2"]

let parse_sources url =
    try
        let dirname = Filename.dirname url in
        let basename = Filename.basename url in
        let extension = List.find (Filename.check_suffix url) extensions in
        let basename' = Filename.chop_suffix basename extension in
        (dirname, basename', extension)
    with
        Not_found ->
            raise (Syntax.Error (Lexing.dummy_pos, Lexing.dummy_pos, "Invalid extension in `Sources:' line"))

(*
let parse_category = function
    | "godi" -> `godi
    | "apps" -> `apps
    | "conf" -> `conf
    | s -> raise (ParseError ("Invalid category in dependency: " ^ s))

let parse_package package (* with category *) =
    try
        let dash_pos = String.index package '-' in
        (parse_category (String.sub package 0 dash_pos),
         String.sub package (dash_pos + 1)
                            (String.length package - dash_pos - 1))
    with
        Not_found ->
            raise (ParseError ("Invalid dependency -- no dash"))

let parse_option = function
    | "configure" -> `configure
    | "bmake" -> `bmake
    | "opt" -> `opt
    | "htdoc" -> `htdoc
    | s -> raise (ParseError ("Invalid option in `Options:' line: " ^ s))
*)
(*
let parse_error s = prerr_endline ("Dame that is a sweet " ^ s);
                    Lexing.
*)

let missing rhs_n s =
    let pos = Parsing.rhs_end_pos rhs_n in
    raise (Syntax.Error (pos, pos, s))

let error n s =
    raise (Syntax.Error (Parsing.rhs_start_pos n, Parsing.rhs_end_pos n, s))
%}

%token PACKAGE
%token VERSION
%token REVISION
%token DEPENDS BUILD_DEPENDS
%token SOURCES
%token HOMEPAGE
%token MAINTAINER
%token DESCRIPTION
%token OPTIONS
%token DOCFILES
%token <string> STRING
%token <string * string> STRINGPAIR
%token <int> INT
%token <Control.relop> RELOP
%token <Control.options> OPTION
%token LPAREN RPAREN COMMA
%token EOF
%token <Control.category * string> PKGNAME

%start spec
%type <Control.t> spec

%%

/* idea
spec:
    package_line
    version_line
    ...
(since order is already checked by lexer)
*/

spec:
    | PACKAGE PKGNAME version_etc { $3 $2 }
    | PACKAGE error version_etc { missing 1 "No package name?" }
    | PACKAGE PKGNAME error { error 3 "Trailing junk after package name?" }
;

version_etc:
    | VERSION STRING revision_etc { $3 $2 }
    | VERSION error { missing 1 "No version number?" }
    | VERSION STRING error { error 3 "Trailing junk after version number?" }
;

revision_etc:
    | REVISION INT depends_etc  { $3 $2 }
    | REVISION error { missing 1 "No revision number?" }
    | REVISION INT error { error 3 "Trailing junk after revision number?" }
;

depends_etc:
    | DEPENDS deps build_depends_etc { $3 $2 }
    /* this should probably be handled in deps below... */
    /* | DEPENDS error { error 2 "Error in dependencies." } */
    | DEPENDS deps error { error 3 "Trailing junk after dependencies." }
;

build_depends_etc:
    | BUILD_DEPENDS deps sources_etc { $3 $2 }
    /* this should probably be handled in deps below... */
    /* | BUILD_DEPENDS error { error 2 "Error in build dependencies." } */
    | BUILD_DEPENDS deps error { error 3 "Trailing junk after build dependencies." }
;

sources_etc:
    | SOURCES STRING homepage_etc { $3 (parse_sources $2) }
    | SOURCES error { missing 1 "No sources url?" }
    | SOURCES STRING error { error 3 "Trailing junk after sources url?" }
;

homepage_etc:
    | HOMEPAGE STRING maintainer_etc { $3 $2 }
    | HOMEPAGE error { missing 1 "No homepage url?" }
    | HOMEPAGE STRING error { error 3 "Trailing junk after homepage url?" }
;

maintainer_etc:
    | MAINTAINER STRING description_etc { $3 $2 }
    | MAINTAINER error { missing 1 "No maintainer string?" }
    /* trailing junk impossible: maintainer goes til end of line */
;

description_etc:
    | DESCRIPTION STRINGPAIR options_etc { $3 $2 }
;

options_etc:
    | OPTIONS options docfiles_etc { $3 $2 }
    /* XXX this should probably be handled in options, below... */
    /* | OPTIONS error { error 2 "Error in options." } */
    | OPTIONS options error { error 3 "Trailing junk after options?" }
;

docfiles_etc:
    | DOCFILES docfiles eof { $3 $2 }
    /* XXX this should probably be handled in docfiles, below... */
    /* | DOCFILES error { error 2 "Error in docfiles." } */
    | DOCFILES docfiles error { error 3 "Trailing junk after docfiles?" }
;

eof:
    | EOF {
            fun docfiles options (short_desc, long_desc) maintainer homepage
                (sources_site, sources_basename, sources_extension)
                build_deps deps revision version (category, name) ->
                    {
                        package = name;
                        category = category;
                        version = version;
                        revision = revision;
                        depends = deps;
                        build_depends = build_deps;
                        sources_site = sources_site;
                        sources_basename = sources_basename;
                        sources_extension = sources_extension;
                        homepage = homepage;
                        maintainer = maintainer;
                        short_desc = short_desc;
                        long_desc = long_desc;
                        options = options;
                        docfiles = docfiles;
                    }
            }

deps:
  | dep COMMA deps  { $1 :: $3 }
  | dep             { [$1] }
  | /* epsilon */   { [] }
;

dep:
  | PKGNAME LPAREN RELOP STRING RPAREN
        { let (cat, name) = $1 in (cat, name, Some ($3, $4)) }
  | PKGNAME
        { let (cat, name) = $1 in (cat, name, None) }
;

options:
  | OPTION COMMA options    { $1 :: $3 }
  | OPTION                  { [$1] }
  | /* epsilon */           { [] }
;

docfiles:
  | STRING COMMA docfiles   { $1 :: $3 }
  | STRING                  { [$1] }
  | /* epsilon */           { [] }
;
